import time
import enum
from typing import Any
from ophyd import Device, Component, EpicsSignal, EpicsSignalRO, Kind
from ophyd import PVPositioner, Signal, DeviceStatus
from ophyd.pseudopos import (
    pseudo_position_argument,
    real_position_argument,
    PseudoSingle,
    PseudoPositioner,
)
from ophyd.device import Staged

from ophyd_devices.epics.devices.bec_scaninfo_mixin import BecScaninfoMixin
from ophyd_devices.utils import bec_utils

from bec_lib import bec_logger


logger = bec_logger.logger


class DelayGeneratorError(Exception):
    """Exception raised for errors."""


class DeviceInitError(DelayGeneratorError):
    """Error upon failed initialization, invoked by missing device manager or device not started in sim_mode."""


class DelayGeneratorNotOkay(DelayGeneratorError):
    """Error when DDG is not okay"""


class TriggerSource(enum.IntEnum):
    """
    Class for trigger options of DG645

    Used to set the trigger source of the DG645 by setting the value
    e.g. source.put(TriggerSource.Internal)
    Exp:
        TriggerSource.Internal
    """

    INTERNAL = 0
    EXT_RISING_EDGE = 1
    EXT_FALLING_EDGE = 2
    SS_EXT_RISING_EDGE = 3
    SS_EXT_FALLING_EDGE = 4
    SINGLE_SHOT = 5
    LINE = 6


class DelayStatic(Device):
    """
    Static axis for the T0 output channel

    It allows setting the logic levels, but the timing is fixed.
    The signal is high after receiving the trigger until the end
    of the holdoff period.
    """

    # Other channel stuff
    ttl_mode = Component(EpicsSignal, "OutputModeTtlSS.PROC", kind=Kind.config)
    nim_mode = Component(EpicsSignal, "OutputModeNimSS.PROC", kind=Kind.config)
    polarity = Component(
        EpicsSignal,
        "OutputPolarityBI",
        write_pv="OutputPolarityBO",
        name="polarity",
        kind=Kind.config,
    )
    amplitude = Component(
        EpicsSignal,
        "OutputAmpAI",
        write_pv="OutputAmpAO",
        name="amplitude",
        kind=Kind.config,
    )
    offset = Component(
        EpicsSignal,
        "OutputOffsetAI",
        write_pv="OutputOffsetAO",
        name="offset",
        kind=Kind.config,
    )


class DummyPositioner(PVPositioner):
    """Dummy Positioner to set AO, AI and ReferenceMO."""

    setpoint = Component(EpicsSignal, "DelayAO", put_complete=True, kind=Kind.config)
    readback = Component(EpicsSignalRO, "DelayAI", kind=Kind.config)
    done = Component(Signal, value=1)
    reference = Component(EpicsSignal, "ReferenceMO", put_complete=True, kind=Kind.config)


class DelayPair(PseudoPositioner):
    """
    Delay pair interface

    Virtual motor interface to a pair of signals (on the frontpanel - AB/CD/EF/GH).
    It offers a simple delay and pulse width interface.
    """

    # The pseudo positioner axes
    delay = Component(PseudoSingle, limits=(0, 2000.0), name="delay")
    width = Component(PseudoSingle, limits=(0, 2000.0), name="pulsewidth")
    ch1 = Component(DummyPositioner, name="ch1")
    ch2 = Component(DummyPositioner, name="ch2")
    io = Component(DelayStatic, name="io")

    def __init__(self, *args, **kwargs):
        # Change suffix names before connecting (a bit of dynamic connections)
        self.__class__.__dict__["ch1"].suffix = kwargs["channel"][0]
        self.__class__.__dict__["ch2"].suffix = kwargs["channel"][1]
        self.__class__.__dict__["io"].suffix = kwargs["channel"]

        del kwargs["channel"]
        # Call parent to start the connections
        super().__init__(*args, **kwargs)

    @pseudo_position_argument
    def forward(self, pseudo_pos):
        """Run a forward (pseudo -> real) calculation"""
        return self.RealPosition(ch1=pseudo_pos.delay, ch2=pseudo_pos.delay + pseudo_pos.width)

    @real_position_argument
    def inverse(self, real_pos):
        """Run an inverse (real -> pseudo) calculation"""
        return self.PseudoPosition(delay=real_pos.ch1, width=real_pos.ch2 - real_pos.ch1)


class DDGCustomMixin:
    """
    Mixin class for custom DelayGenerator logic within PSIDelayGeneratorBase.

    This class provides a parent class for implementation of BL specific logic of the device.
    It is also possible to pass implementing certain methods, e.g. finished or on_trigger,
    based on the setup and desired operation mode at the beamline.

    Args:
        parent (object): instance of PSIDelayGeneratorBase
        **kwargs: keyword arguments
    """

    def __init__(self, *_args, parent: Device = None, **_kwargs) -> None:
        self.parent = parent

    def initialize_default_parameter(self) -> None:
        """
        Method to initialize default parameters for DDG.

        Called upon initiating the base class.
        It should be used to set the DDG default parameters.
        These may include: amplitude, offsets, delays, etc.
        """

    def prepare_ddg(self) -> None:
        """
        Method to prepare the DDG for the upcoming scan.

        Called by the stage method of the base class.
        It should be used to set the DDG parameters for the upcoming scan.
        """

    def on_trigger(self) -> None:
        """Method executed upon trigger call in parent class"""

    def finished(self) -> None:
        """Method to check if DDG is finished with the scan"""

    def on_pre_scan(self) -> None:
        """
        Method executed upon pre_scan call in parent class.

        Covenient to implement time sensitive actions to be executed right before start of the scan.
        Example could be to open the shutter by triggering a pulse via pre_scan.
        """

    def check_scanID(self) -> None:
        """Method to check if there is a new scanID, called by stage."""

    def is_ddg_okay(self, raise_on_error=False) -> None:
        """
        Method to check if DDG is okay

        It checks the status PV of the DDG and tries to clear the error if it is not okay.
        It will rerun itself and raise DelayGeneratorNotOkay if DDG is still not okay.

        Args:
            raise_on_error (bool, optional): raise exception if DDG is not okay. Defaults to False.
        """
        status = self.parent.status.read()[self.parent.status.name]["value"]
        if status != "STATUS OK" and not raise_on_error:
            logger.warning(f"DDG returns {status}, trying to clear ERROR")
            self.parent.clear_error()
            time.sleep(1)
            self.is_ddg_okay(raise_on_error=True)
        elif status != "STATUS OK":
            raise DelayGeneratorNotOkay(f"DDG failed to start with status: {status}")


class PSIDelayGeneratorBase(Device):
    """
    Abstract base class for DelayGenerator DG645

    This class implements a thin Ophyd wrapper around the Stanford Research DG645
    digital delay generator.

    The DG645 generates 8+1 signals: A, B, C, D, E, F, G, H and T0. Front panel outputs
    T0, AB, CD, EF and GH are combinations of these signals. Back panel outputs are
    directly routed signals. Signals are not independent.

    Signal pairs, e.g. AB, CD, EF, GH, are implemented as DelayPair objects. They
    have a TTL pulse width, delay and a reference signal to which they are being triggered.
    In addition, the io layer allows setting amplitude, offset and polarity for each pair.

    Detailed information can be found in the manual:
    https://www.thinksrs.com/downloads/pdfs/manuals/DG645m.pdf

    Class attributes:
        custom_prepare_cls (object): class for custom prepare logic (BL specific)

    Args:
        prefix (str)                : EPICS PV prefix for component (optional)
        name (str)                  : name of the device, as will be reported via read()
        kind (str)                  : member of class 'ophydobj.Kind', defaults to Kind.normal
                                        omitted -> readout ignored for read 'ophydobj.read()'
                                        normal -> readout for read
                                        config -> config parameter for 'ophydobj.read_configuration()'
                                        hinted -> which attribute is readout for read
        read_attrs (list)           : sequence of attribute names to read
        configuration_attrs (list)  : sequence of attribute names via config_parameters
        parent (object)             : instance of the parent device
        device_manager (object)     : bec device manager
        sim_mode (bool)             : simulation mode, if True, no device manager is required
        **kwargs                    : keyword arguments
        attributes                  : lazy_wait_for_connection : bool
    """

    # Custom_prepare_cls
    custom_prepare_cls = DDGCustomMixin

    SUB_PROGRESS = "progress"
    SUB_VALUE = "value"
    _default_sub = SUB_VALUE

    USER_ACCESS = [
        "set_channels",
        "_set_trigger",
        "burst_enable",
        "burst_disable",
        "reload_config",
    ]

    # Assign PVs from DDG645
    trigger_burst_readout = Component(
        EpicsSignal, "EventStatusLI.PROC", name="trigger_burst_readout"
    )
    burst_cycle_finished = Component(EpicsSignalRO, "EventStatusMBBID.B3", name="read_burst_state")
    delay_finished = Component(EpicsSignalRO, "EventStatusMBBID.B2", name="delay_finished")
    status = Component(EpicsSignalRO, "StatusSI", name="status")
    clear_error = Component(EpicsSignal, "StatusClearBO", name="clear_error")

    # Front Panel
    channelT0 = Component(DelayStatic, "T0", name="T0")
    channelAB = Component(DelayPair, "", name="AB", channel="AB")
    channelCD = Component(DelayPair, "", name="CD", channel="CD")
    channelEF = Component(DelayPair, "", name="EF", channel="EF")
    channelGH = Component(DelayPair, "", name="GH", channel="GH")

    holdoff = Component(
        EpicsSignal,
        "TriggerHoldoffAI",
        write_pv="TriggerHoldoffAO",
        name="trigger_holdoff",
        kind=Kind.config,
    )
    inhibit = Component(
        EpicsSignal,
        "TriggerInhibitMI",
        write_pv="TriggerInhibitMO",
        name="trigger_inhibit",
        kind=Kind.config,
    )
    source = Component(
        EpicsSignal,
        "TriggerSourceMI",
        write_pv="TriggerSourceMO",
        name="trigger_source",
        kind=Kind.config,
    )
    level = Component(
        EpicsSignal,
        "TriggerLevelAI",
        write_pv="TriggerLevelAO",
        name="trigger_level",
        kind=Kind.config,
    )
    rate = Component(
        EpicsSignal,
        "TriggerRateAI",
        write_pv="TriggerRateAO",
        name="trigger_rate",
        kind=Kind.config,
    )
    trigger_shot = Component(EpicsSignal, "TriggerDelayBO", name="trigger_shot", kind="config")
    burstMode = Component(
        EpicsSignal,
        "BurstModeBI",
        write_pv="BurstModeBO",
        name="burstmode",
        kind=Kind.config,
    )
    burstConfig = Component(
        EpicsSignal,
        "BurstConfigBI",
        write_pv="BurstConfigBO",
        name="burstconfig",
        kind=Kind.config,
    )
    burstCount = Component(
        EpicsSignal,
        "BurstCountLI",
        write_pv="BurstCountLO",
        name="burstcount",
        kind=Kind.config,
    )
    burstDelay = Component(
        EpicsSignal,
        "BurstDelayAI",
        write_pv="BurstDelayAO",
        name="burstdelay",
        kind=Kind.config,
    )
    burstPeriod = Component(
        EpicsSignal,
        "BurstPeriodAI",
        write_pv="BurstPeriodAO",
        name="burstperiod",
        kind=Kind.config,
    )

    def __init__(
        self,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        device_manager=None,
        sim_mode=False,
        **kwargs,
    ):
        super().__init__(
            prefix=prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs,
        )
        if device_manager is None and not sim_mode:
            raise DeviceInitError(
                f"No device manager for device: {name}, and not started sim_mode: {sim_mode}. Add"
                " DeviceManager to initialization or init with sim_mode=True"
            )
        # Init variables
        self.sim_mode = sim_mode
        self.stopped = False
        self.name = name
        self.scaninfo = None
        self.timeout = 5
        self.all_channels = [
            "channelT0",
            "channelAB",
            "channelCD",
            "channelEF",
            "channelGH",
        ]
        self.all_delay_pairs = ["AB", "CD", "EF", "GH"]
        self.wait_for_connection(all_signals=True)

        # Init custom prepare class with BL specific logic
        self.custom_prepare = self.custom_prepare_cls(parent=self, **kwargs)
        if not sim_mode:
            self.device_manager = device_manager
        else:
            self.device_manager = bec_utils.DMMock()
        self.producer = self.device_manager.producer
        self._update_scaninfo()
        self._init()

    def _update_scaninfo(self) -> None:
        """
        Method to updated scaninfo from BEC.

        In sim_mode, scaninfo output is mocked - see bec_scaninfo_mixin.py
        """
        self.scaninfo = BecScaninfoMixin(self.device_manager, self.sim_mode)
        self.scaninfo.load_scan_metadata()

    def _init(self) -> None:
        """Method to initialize custom parameters of the DDG."""
        self.custom_prepare.initialize_default_parameter()
        self.custom_prepare.is_ddg_okay()

    def set_channels(self, signal: str, value: Any, channels: list = None) -> None:
        """
        Method to set signals on DelayPair and DelayStatic channels.

        Signals can be set on the DelayPair and DelayStatic channels. The method checks
        if the signal is available on the channel and sets it. It works for both, DelayPair
        and Delay Static although signals are hosted in different layers.

        Args:
            signal (str)                : signal to set (width, delay, amplitude, offset, polarity)
            value (Any)                 : value to set
            channels (list, optional)   : list of channels to set. Defaults to self.all_channels (T0,AB,CD,EF,GH)
        """
        if not channels:
            channels = self.all_channels
        for chname in channels:
            channel = getattr(self, chname, None)
            if not channel:
                continue
            if signal in channel.component_names:
                getattr(channel, signal).set(value)
                continue
            if "io" in channel.component_names and signal in channel.io.component_names:
                getattr(channel.io, signal).set(value)

    def set_trigger(self, trigger_source: TriggerSource) -> None:
        """Set trigger source on DDG - possible values defined in TriggerSource enum"""
        value = int(trigger_source)
        self.source.put(value)

    def burst_enable(self, count, delay, period, config="all"):
        """Enable the burst mode"""
        # Validate inputs
        count = int(count)
        assert count > 0, "Number of bursts must be positive"
        assert delay >= 0, "Burst delay must be larger than 0"
        assert period > 0, "Burst period must be positive"
        assert config in [
            "all",
            "first",
        ], "Supported burst configs are 'all' and 'first'"

        self.burstMode.put(1)
        self.burstCount.put(count)
        self.burstDelay.put(delay)
        self.burstPeriod.put(period)

        if config == "all":
            self.burstConfig.put(0)
        elif config == "first":
            self.burstConfig.put(1)

    def burst_disable(self):
        """Disable burst mode"""
        self.burstMode.put(0)

    def stage(self) -> list[object]:
        """
        Method to stage the device.

        Called in preparation for a scan.

        Internal Calls:
        - scaninfo.load_scan_metadata        : load scan metadata
        - custom_prepare.prepare_ddg         : prepare DDG for measurement
        - is_ddg_okay                        : check if DDG is okay

        Returns:
            list(object): list of objects that were staged
        """
        if self._staged != Staged.no:
            return super().stage()
        self.stopped = False
        self.scaninfo.load_scan_metadata()
        self.custom_prepare.prepare_ddg()
        self.custom_prepare.is_ddg_okay()
        # At the moment needed bc signal might not be reliable, BEC too fast.
        # Consider removing this overhead in future!
        time.sleep(0.05)
        return super().stage()

    def trigger(self) -> DeviceStatus:
        """
        Method to trigger the acquisition.

        Internal Call:
        - custom_prepare.on_trigger  : execute BL specific action
        """
        self.custom_prepare.on_trigger()
        return super().trigger()

    def pre_scan(self) -> None:
        """
        Method pre_scan gets executed directly before the scan

        Internal Call:
        - custom_prepare.on_pre_scan  : execute BL specific action
        """
        self.custom_prepare.on_pre_scan()

    def unstage(self) -> list[object]:
        """
        Method unstage gets called at the end of a scan.

        If scan (self.stopped is True) is stopped, returns directly.
        Otherwise, checks if the DDG finished acquisition

        Internal Calls:
        - custom_prepare.check_scanID          : check if scanID changed or detector stopped
        - custom_prepare.finished              : check if device finished acquisition (succesfully)
        - is_ddg_okay                          : check if DDG is okay

        Returns:
            list(object): list of objects that were unstaged
        """
        self.custom_prepare.check_scanID()
        if self.stopped is True:
            return super().unstage()
        self.custom_prepare.finished()
        self.custom_prepare.is_ddg_okay()
        self.stopped = False
        return super().unstage()

    def stop(self, *, success=False) -> None:
        """
        Method to stop the DDG

        #TODO Check if the pulse generation can be interruppted

        Internal Call:
        - custom_prepare.is_ddg_okay          : check if DDG is okay
        """
        self.custom_prepare.is_ddg_okay()
        super().stop(success=success)
        self.stopped = True



import numpy as np
from ophyd import Component, Device, EpicsSignal, EpicsSignalRO


class SpmBase(Device):
    """Python wrapper for the Staggered Blade Pair Monitors

    SPM's consist of a set of four horizontal tungsten blades and are
    used to monitor the beam height (only Y) for the bending magnet
    beamlines of SLS.

    Note: EPICS provided signals are read only, but the users can
    change the beam position offset.
    """

    # Motor interface
    s1 = Component(EpicsSignalRO, "Current1", auto_monitor=True)
    s2 = Component(EpicsSignalRO, "Current2", auto_monitor=True)
    s3 = Component(EpicsSignalRO, "Current3", auto_monitor=True)
    s4 = Component(EpicsSignalRO, "Current4", auto_monitor=True)
    sum = Component(EpicsSignalRO, "SumAll", auto_monitor=True)
    y = Component(EpicsSignalRO, "Y", auto_monitor=True)
    scale = Component(EpicsSignal, "PositionScaleY", auto_monitor=True)
    offset = Component(EpicsSignal, "PositionOffsetY", auto_monitor=True)


class SpmSim(SpmBase):
    """Python wrapper for simulated Staggered Blade Pair Monitors

    SPM's consist of a set of four horizontal tungsten blades and are
    used to monitor the beam height (only Y) for the bending magnet
    beamlines of SLS.

    This simulation device extends the basic proxy with a script that
    fills signals with quasi-randomized values.
    """

    # Motor interface
    s1w = Component(EpicsSignal, "Current1:RAW.VAL", auto_monitor=False)
    s2w = Component(EpicsSignal, "Current2:RAW.VAL", auto_monitor=False)
    s3w = Component(EpicsSignal, "Current3:RAW.VAL", auto_monitor=False)
    s4w = Component(EpicsSignal, "Current4:RAW.VAL", auto_monitor=False)
    rangew = Component(EpicsSignal, "RANGEraw", auto_monitor=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._MX = 0
        self._MY = 0
        self._I0 = 255.0
        self._x = np.linspace(-5, 5, 64)
        self._y = np.linspace(-5, 5, 64)
        self._x, self._y = np.meshgrid(self._x, self._y)

    def _simFrame(self):
        """Generator to simulate a jumping gaussian"""

        # Define normalized 2D gaussian
        def gaus2d(x=0, y=0, mx=0, my=0, sx=1, sy=1):
            return np.exp(
                -((x - mx) ** 2.0 / (2.0 * sx**2.0) + (y - my) ** 2.0 / (2.0 * sy**2.0))
            )

        # Generator for dynamic values
        self._MX = 0.75 * self._MX + 0.25 * (10.0 * np.random.random() - 5.0)
        self._MY = 0.75 * self._MY + 0.25 * (10.0 * np.random.random() - 5.0)
        self._I0 = 0.75 * self._I0 + 0.25 * (255.0 * np.random.random())

        arr = self._I0 * gaus2d(self._x, self._y, self._MX, self._MY)
        return arr

    def sim(self):
        # Get next frame
        beam = self._simFrame()
        total = np.sum(beam) - np.sum(beam[24:48, :])
        rnge = np.floor(np.log10(total) - 0.0)
        s1 = np.sum(beam[0:16, :]) / 10**rnge
        s2 = np.sum(beam[16:24, :]) / 10**rnge
        s3 = np.sum(beam[40:48, :]) / 10**rnge
        s4 = np.sum(beam[48:64, :]) / 10**rnge

        self.s1w.set(s1).wait()
        self.s2w.set(s2).wait()
        self.s3w.set(s3).wait()
        self.s4w.set(s4).wait()
        self.rangew.set(rnge).wait()
        # Print debug info
        print(f"Raw signals: R={rnge}\t{s1}\t{s2}\t{s3}\t{s4}")
        # plt.imshow(beam)
        # plt.show(block=False)
        # plt.pause(0.5)


# Automatically start simulation if directly invoked
if __name__ == "__main__":
    spm1 = SpmSim("X06D-FE-BM1:", name="spm1")
    spm2 = SpmSim("X06D-FE-BM2:", name="spm2")

    spm1.wait_for_connection(timeout=5)
    spm2.wait_for_connection(timeout=5)

    spm1.rangew.set(1).wait()
    spm2.rangew.set(1).wait()

    while True:
        print("---")
        spm1.sim()
        spm2.sim()



from .sim import (
    SynAxisMonitor,
    SimPositioner,
    SynDynamicComponents,
    SynFlyer,
    SynSignalRO,
    SynSLSDetector,
)
from .sim_xtreme import SynXtremeOtf

from .sim_signals import ReadbackSignal, SetpointSignal, IsMovingSignal



import functools
import threading
import time

import numpy as np
from bec_lib import messages, MessageEndpoints, bec_logger
from ophyd import Component as Cpt
from ophyd import Device, PositionerBase, Signal
from ophyd.status import wait as status_wait
from ophyd.utils import LimitError, ReadOnlyError

from ophyd_devices.utils.controller import Controller, threadlocked
from ophyd_devices.utils.socket import SocketIO, SocketSignal, raise_if_disconnected

logger = bec_logger.logger


class RtLamniCommunicationError(Exception):
    pass


class RtLamniError(Exception):
    pass


class BECConfigError(Exception):
    pass


def retry_once(fcn):
    """Decorator to rerun a function in case a CommunicationError was raised. This may happen if the buffer was not empty."""

    @functools.wraps(fcn)
    def wrapper(self, *args, **kwargs):
        try:
            val = fcn(self, *args, **kwargs)
        except (RtLamniCommunicationError, RtLamniError):
            val = fcn(self, *args, **kwargs)
        return val

    return wrapper


class RtLamniController(Controller):
    USER_ACCESS = [
        "socket_put_and_receive",
        "set_rotation_angle",
        "feedback_disable",
        "feedback_enable_without_reset",
        "feedback_disable_and_even_reset_lamni_angle_interferometer",
        "feedback_enable_with_reset",
        "add_pos_to_scan",
        "clear_trajectory_generator",
        "_set_axis_velocity",
        "_set_axis_velocity_maximum_speed",
        "_position_sampling_single_read",
        "_position_sampling_single_reset_and_start_sampling",
    ]

    def __init__(
        self,
        *,
        name="RtLamniController",
        kind=None,
        parent=None,
        socket=None,
        attr_name="",
        labels=None,
    ):
        if not hasattr(self, "_initialized") or not self._initialized:
            self._rtlamni_axis_per_controller = 3
            self._axis = [None for axis_num in range(self._rtlamni_axis_per_controller)]
            self._min_scan_buffer_reached = False
            super().__init__(
                name=name,
                socket=socket,
                attr_name=attr_name,
                parent=parent,
                labels=labels,
                kind=kind,
            )
        self.readout_metadata = {}

    def on(self, controller_num=0) -> None:
        """Open a new socket connection to the controller"""
        if not self.connected:
            try:
                self.sock.open()
                # discuss - after disconnect takes a while for the server to be ready again
                max_retries = 10
                tries = 0
                while not self.connected:
                    try:
                        welcome_message = self.sock.receive()
                        self.connected = True
                    except ConnectionResetError as conn_reset:
                        if tries > max_retries:
                            raise conn_reset
                        tries += 1
                        time.sleep(2)
            except ConnectionRefusedError as conn_error:
                logger.error("Failed to open a connection to RTLamNI.")
                raise RtLamniCommunicationError from conn_error

        else:
            logger.info("The connection has already been established.")
            # warnings.warn(f"The connection has already been established.", stacklevel=2)

        self._update_flyer_device_info()

    def off(self) -> None:
        """Close the socket connection to the controller"""
        if self.connected:
            self.sock.close()
            self.connected = False
        else:
            logger.info("The connection is already closed.")

    def set_axis(self, axis: Device, axis_nr: int) -> None:
        """Assign an axis to a device instance.

        Args:
            axis (Device): Device instance (e.g. GalilMotor)
            axis_nr (int): Controller axis number

        """
        self._axis[axis_nr] = axis

    @threadlocked
    def socket_put(self, val: str) -> None:
        self.sock.put(f"{val}\n".encode())

    @threadlocked
    def socket_get(self) -> str:
        return self.sock.receive().decode()

    @retry_once
    @threadlocked
    def socket_put_and_receive(self, val: str, remove_trailing_chars=True) -> str:
        self.socket_put(val)
        if remove_trailing_chars:
            return self._remove_trailing_characters(self.sock.receive().decode())
        return self.socket_get()

    def is_axis_moving(self, axis_Id) -> bool:
        # this checks that axis is on target
        axis_is_on_target = bool(float(self.socket_put_and_receive(f"o")))
        return not axis_is_on_target

    #    def is_thread_active(self, thread_id: int) -> bool:
    #        val = float(self.socket_put_and_receive(f"MG_XQ{thread_id}"))
    #        if val == -1:
    #            return False
    #        return True

    def _remove_trailing_characters(self, var) -> str:
        if len(var) > 1:
            return var.split("\r\n")[0]
        return var

    @threadlocked
    def set_rotation_angle(self, val: float):
        self.socket_put(f"a{(val-300+30.538)/180*np.pi}")

    @threadlocked
    def stop_all_axes(self):
        self.socket_put("sc")

    @threadlocked
    def feedback_disable(self):
        self.socket_put("J0")
        logger.info("LamNI Feedback disabled.")
        self.set_device_enabled("lsamx", True)
        self.set_device_enabled("lsamy", True)
        self.set_device_enabled("loptx", True)
        self.set_device_enabled("lopty", True)
        self.set_device_enabled("loptz", True)

    @threadlocked
    def _set_axis_velocity(self, um_per_s):
        self.socket_put(f"V{um_per_s}")

    @threadlocked
    def _set_axis_velocity_maximum_speed(self):
        self.socket_put(f"V0")

    # for developement of soft continuous scanning
    @threadlocked
    def _position_sampling_single_reset_and_start_sampling(self):
        self.socket_put(f"Ss")

    @threadlocked
    def _position_sampling_single_read(self):
        (number_of_samples, sum0, sum0_2, sum1, sum1_2, sum2, sum2_2) = self.socket_put_and_receive(
            f"Sr"
        ).split(",")
        avg_x = float(sum1) / int(number_of_samples)
        avg_y = float(sum0) / int(number_of_samples)
        stdev_x = np.sqrt(
            float(sum1_2) / int(number_of_samples)
            - np.power(float(sum1) / int(number_of_samples), 2)
        )
        stdev_y = np.sqrt(
            float(sum0_2) / int(number_of_samples)
            - np.power(float(sum0) / int(number_of_samples), 2)
        )
        return (avg_x, avg_y, stdev_x, stdev_y)

    @threadlocked
    def feedback_enable_without_reset(self):
        # read current interferometer position
        return_table = (self.socket_put_and_receive(f"J4")).split(",")
        x_curr = float(return_table[2])
        y_curr = float(return_table[1])
        # set these as closed loop target position
        self.socket_put(f"pa0,{x_curr:.4f}")
        self.socket_put(f"pa1,{y_curr:.4f}")
        self.get_device_manager().devices.rtx.obj.user_setpoint.set_with_feedback_disabled(x_curr)
        self.get_device_manager().devices.rty.obj.user_setpoint.set_with_feedback_disabled(y_curr)
        self.socket_put("J5")
        logger.info("LamNI Feedback enabled (without reset).")
        self.set_device_enabled("lsamx", False)
        self.set_device_enabled("lsamy", False)
        self.set_device_enabled("loptx", False)
        self.set_device_enabled("lopty", False)
        self.set_device_enabled("loptz", False)

    @threadlocked
    def feedback_disable_and_even_reset_lamni_angle_interferometer(self):
        self.socket_put("J6")
        logger.info("LamNI Feedback disabled including the angular interferometer.")
        self.set_device_enabled("lsamx", True)
        self.set_device_enabled("lsamy", True)
        self.set_device_enabled("loptx", True)
        self.set_device_enabled("lopty", True)
        self.set_device_enabled("loptz", True)

    def get_device_manager(self):
        for axis in self._axis:
            if hasattr(axis, "device_manager") and axis.device_manager:
                return axis.device_manager
        raise BECConfigError("Could not access the device_manager")

    def get_axis_by_name(self, name):
        for axis in self._axis:
            if axis:
                if axis.name == name:
                    return axis
        raise RuntimeError(f"Could not find an axis with name {name}")

    @threadlocked
    def clear_trajectory_generator(self):
        self.socket_put("sc")
        logger.info("LamNI scan stopped and deleted, moving to start position")

    def add_pos_to_scan(self, positions) -> None:
        def send_positions(parent, positions):
            parent._min_scan_buffer_reached = False
            for pos_index, pos in enumerate(positions):
                parent.socket_put_and_receive(f"s{pos[0]},{pos[1]},0")
                if pos_index > 100:
                    parent._min_scan_buffer_reached = True
            parent._min_scan_buffer_reached = True

        threading.Thread(target=send_positions, args=(self, positions), daemon=True).start()

    @retry_once
    @threadlocked
    def get_scan_status(self):
        return_table = (self.socket_put_and_receive(f"sr")).split(",")
        if len(return_table) != 3:
            raise RtLamniCommunicationError(
                f"Expected to receive 3 return values. Instead received {return_table}"
            )
        mode = int(return_table[0])
        # mode 0: direct positioning
        # mode 1: running internal timer (not tested/used anymore)
        # mode 2: rt point scan running
        # mode 3: rt point scan starting
        # mode 5/6: rt continuous scanning (not available in LamNI)
        number_of_positions_planned = int(return_table[1])
        current_position_in_scan = int(return_table[2])
        return (mode, number_of_positions_planned, current_position_in_scan)

    @threadlocked
    def start_scan(self):
        interferometer_feedback_not_running = int((self.socket_put_and_receive("J2")).split(",")[0])
        if interferometer_feedback_not_running == 1:
            logger.error(
                "Cannot start scan because feedback loop is not running or there is an interferometer error."
            )
            raise RtLamniError(
                "Cannot start scan because feedback loop is not running or there is an interferometer error."
            )
            # here exception
        (mode, number_of_positions_planned, current_position_in_scan) = self.get_scan_status()

        if number_of_positions_planned == 0:
            logger.error("Cannot start scan because no target positions are planned.")
            raise RtLamniError("Cannot start scan because no target positions are planned.")
            # hier exception
        # start a point-by-point scan (for cont scan in flomni it would be "sa")
        self.socket_put_and_receive("sd")

    def start_readout(self):
        readout = threading.Thread(target=self.read_positions_from_sampler)
        readout.start()

    def _update_flyer_device_info(self):
        flyer_info = self._get_flyer_device_info()
        self.get_device_manager().producer.set(
            MessageEndpoints.device_info("rt_scan"),
            messages.DeviceInfoMessage(device="rt_scan", info=flyer_info).dumps(),
        )

    def _get_flyer_device_info(self) -> dict:
        return {
            "device_name": self.name,
            "device_attr_name": getattr(self, "attr_name", ""),
            "device_dotted_name": getattr(self, "dotted_name", ""),
            "device_info": {
                "device_base_class": "ophydobject",
                "signals": [],
                "hints": {"fields": ["average_x_st_fzp", "average_y_st_fzp"]},
                "describe": {},
                "describe_configuration": {},
                "sub_devices": [],
                "custom_user_access": [],
            },
        }

    def kickoff(self, metadata):
        self.readout_metadata = metadata
        while not self._min_scan_buffer_reached:
            time.sleep(0.001)
        self.start_scan()
        time.sleep(0.1)
        self.start_readout()

    def _get_signals_from_table(self, return_table) -> dict:
        self.average_stdeviations_x_st_fzp += float(return_table[5])
        self.average_stdeviations_y_st_fzp += float(return_table[8])
        self.average_lamni_angle += float(return_table[19])
        signals = {
            "target_x": {"value": float(return_table[3])},
            "average_x_st_fzp": {"value": float(return_table[4])},
            "stdev_x_st_fzp": {"value": float(return_table[5])},
            "target_y": {"value": float(return_table[6])},
            "average_y_st_fzp": {"value": float(return_table[7])},
            "stdev_y_st_fzp": {"value": float(return_table[8])},
            "average_cap1": {"value": float(return_table[9])},
            "stdev_cap1": {"value": float(return_table[10])},
            "average_cap2": {"value": float(return_table[11])},
            "stdev_cap2": {"value": float(return_table[12])},
            "average_cap3": {"value": float(return_table[13])},
            "stdev_cap3": {"value": float(return_table[14])},
            "average_cap4": {"value": float(return_table[15])},
            "stdev_cap4": {"value": float(return_table[16])},
            "average_cap5": {"value": float(return_table[17])},
            "stdev_cap5": {"value": float(return_table[18])},
            "average_angle_interf_ST": {"value": float(return_table[19])},
            "stdev_angle_interf_ST": {"value": float(return_table[20])},
            "average_stdeviations_x_st_fzp": {
                "value": self.average_stdeviations_x_st_fzp / (int(return_table[0]) + 1)
            },
            "average_stdeviations_y_st_fzp": {
                "value": self.average_stdeviations_y_st_fzp / (int(return_table[0]) + 1)
            },
            "average_lamni_angle": {"value": self.average_lamni_angle / (int(return_table[0]) + 1)},
        }
        return signals

    def read_positions_from_sampler(self):
        # this was for reading after the scan completed
        number_of_samples_to_read = 1  # self.get_scan_status()[1]  #number of valid samples, will be updated upon first data read

        read_counter = 0
        previous_point_in_scan = 0

        self.average_stdeviations_x_st_fzp = 0
        self.average_stdeviations_y_st_fzp = 0
        self.average_lamni_angle = 0

        mode, number_of_positions_planned, current_position_in_scan = self.get_scan_status()

        # if not (mode==2 or mode==3):
        #    error
        self.get_device_manager().producer.set_and_publish(
            MessageEndpoints.device_status("rt_scan"),
            messages.DeviceStatusMessage(
                device="rt_scan", status=1, metadata=self.readout_metadata
            ).dumps(),
        )
        # while scan is running
        while mode > 0:
            # logger.info(f"Current scan position {current_position_in_scan} out of {number_of_positions_planned}")
            mode, number_of_positions_planned, current_position_in_scan = self.get_scan_status()
            time.sleep(0.01)
            if current_position_in_scan > 5:
                while current_position_in_scan > read_counter + 1:
                    return_table = (self.socket_put_and_receive(f"r{read_counter}")).split(",")
                    # logger.info(f"{return_table}")
                    logger.info(f"Read {read_counter} out of {number_of_positions_planned}")

                    read_counter = read_counter + 1

                    signals = self._get_signals_from_table(return_table)

                    self.publish_device_data(signals=signals, pointID=int(return_table[0]))

        time.sleep(0.05)

        # read the last samples even though scan is finished already
        while number_of_positions_planned > read_counter:
            return_table = (self.socket_put_and_receive(f"r{read_counter}")).split(",")
            logger.info(f"Read {read_counter} out of {number_of_positions_planned}")
            # logger.info(f"{return_table}")
            read_counter = read_counter + 1

            signals = self._get_signals_from_table(return_table)
            self.publish_device_data(signals=signals, pointID=int(return_table[0]))

        self.get_device_manager().producer.set_and_publish(
            MessageEndpoints.device_status("rt_scan"),
            messages.DeviceStatusMessage(
                device="rt_scan", status=0, metadata=self.readout_metadata
            ).dumps(),
        )

        logger.info(
            f"LamNI statistics: Average of all standard deviations: x {self.average_stdeviations_x_st_fzp/number_of_samples_to_read}, y {self.average_stdeviations_y_st_fzp/number_of_samples_to_read}, angle {self.average_lamni_angle/number_of_samples_to_read}."
        )

    def publish_device_data(self, signals, pointID):
        self.get_device_manager().producer.send(
            MessageEndpoints.device_read("rt_lamni"),
            messages.DeviceMessage(
                signals=signals,
                metadata={"pointID": pointID, **self.readout_metadata},
            ).dumps(),
        )

    def feedback_status_angle_lamni(self) -> bool:
        return_table = (self.socket_put_and_receive(f"J7")).split(",")
        logger.debug(
            f"LamNI angle interferomter status {bool(return_table[0])}, position {float(return_table[1])}, signal {float(return_table[2])}"
        )
        return bool(return_table[0])

    def feedback_enable_with_reset(self):
        if not self.feedback_status_angle_lamni():
            self.feedback_disable_and_even_reset_lamni_angle_interferometer()
            logger.info(f"LamNI resetting interferometer inclusive angular interferomter.")
        else:
            self.feedback_disable()
            logger.info(
                f"LamNI resetting interferomter except angular interferometer which is already running."
            )

        # set these as closed loop target position

        self.socket_put(f"pa0,0")
        self.get_axis_by_name("rtx").user_setpoint.setpoint = 0
        self.socket_put(f"pa1,0")
        self.get_axis_by_name("rty").user_setpoint.setpoint = 0
        self.socket_put(
            f"pa2,0"
        )  # we set all three outputs of the traj. gen. although in LamNI case only 0,1 are used
        self.clear_trajectory_generator()

        self.get_device_manager().devices.lsamrot.obj.move(0, wait=True)

        galil_controller_rt_status = (
            self.get_device_manager().devices.lsamx.obj.controller.lgalil_is_air_off_and_orchestra_enabled()
        )

        if galil_controller_rt_status == 0:
            logger.error(
                "Cannot enable feedback. The small rotation air is on and/or orchestra disabled by the motor controller."
            )
            raise RtLamniError(
                "Cannot enable feedback. The small rotation air is on and/or orchestra disabled by the motor controller."
            )

        time.sleep(0.03)

        lsamx_user_params = self.get_device_manager().devices.lsamx.user_parameter
        if lsamx_user_params is None or lsamx_user_params.get("center") is None:
            raise RuntimeError("lsamx center is not defined")
        lsamy_user_params = self.get_device_manager().devices.lsamy.user_parameter
        if lsamy_user_params is None or lsamy_user_params.get("center") is None:
            raise RuntimeError("lsamy center is not defined")
        lsamx_center = lsamx_user_params.get("center")
        lsamy_center = lsamy_user_params.get("center")
        self.get_device_manager().devices.lsamx.obj.move(lsamx_center, wait=True)
        self.get_device_manager().devices.lsamy.obj.move(lsamy_center, wait=True)
        self.socket_put("J1")

        _waitforfeedbackctr = 0

        interferometer_feedback_not_running = int((self.socket_put_and_receive("J2")).split(",")[0])

        while interferometer_feedback_not_running == 1 and _waitforfeedbackctr < 100:
            time.sleep(0.01)
            _waitforfeedbackctr = _waitforfeedbackctr + 1
            interferometer_feedback_not_running = int(
                (self.socket_put_and_receive("J2")).split(",")[0]
            )

        self.set_device_enabled("lsamx", False)
        self.set_device_enabled("lsamy", False)
        self.set_device_enabled("loptx", False)
        self.set_device_enabled("lopty", False)
        self.set_device_enabled("loptz", False)

        if interferometer_feedback_not_running == 1:
            logger.error(
                "Cannot start scan because feedback loop is not running or there is an interferometer error."
            )
            raise RtLamniError(
                "Cannot start scan because feedback loop is not running or there is an interferometer error."
            )

        time.sleep(0.01)

        # ptychography_alignment_done = 0

    def set_device_enabled(self, device_name: str, enabled: bool) -> None:
        """enable / disable a device"""
        if device_name not in self.get_device_manager().devices:
            logger.warning(
                f"Device {device_name} is not configured and cannot be enabled/disabled."
            )
            return
        self.get_device_manager().devices[device_name].enabled_set = enabled


class RtLamniSignalBase(SocketSignal):
    def __init__(self, signal_name, **kwargs):
        self.signal_name = signal_name
        super().__init__(**kwargs)
        self.controller = self.parent.controller
        self.sock = self.parent.controller.sock


class RtLamniSignalRO(RtLamniSignalBase):
    def __init__(self, signal_name, **kwargs):
        super().__init__(signal_name, **kwargs)
        self._metadata["write_access"] = False

    def _socket_set(self, val):
        raise ReadOnlyError("Read-only signals cannot be set")


class RtLamniReadbackSignal(RtLamniSignalRO):
    @retry_once
    @threadlocked
    def _socket_get(self) -> float:
        """Get command for the readback signal

        Returns:
        float: Readback value after adjusting for sign and motor resolution.
        """
        return_table = (self.controller.socket_put_and_receive(f"J4")).split(",")
        print(return_table)
        if self.parent.axis_Id_numeric == 0:
            readback_index = 2
        elif self.parent.axis_Id_numeric == 1:
            readback_index = 1
        else:
            raise RtLamniError("Currently, only two axes are supported.")

        current_pos = float(return_table[readback_index])

        current_pos *= self.parent.sign
        return current_pos


class RtLamniSetpointSignal(RtLamniSignalBase):
    setpoint = 0

    def _socket_get(self) -> float:
        """Get command for receiving the setpoint / target value.
        The value is not pulled from the controller but instead just the last setpoint used.

        Returns:
            float: setpoint / target value
        """
        return self.setpoint

    @retry_once
    @threadlocked
    def _socket_set(self, val: float) -> None:
        """Set a new target value / setpoint value. Before submission, the target value is adjusted for the axis' sign.
        Furthermore, it is ensured that all axes are referenced before a new setpoint is submitted.

        Args:
            val (float): Target value / setpoint value

        Raises:
            RtLamniError: Raised if interferometer feedback is disabled.

        """
        interferometer_feedback_not_running = int(
            (self.controller.socket_put_and_receive("J2")).split(",")[0]
        )
        if interferometer_feedback_not_running != 0:
            raise RtLamniError(
                "The interferometer feedback is not running. Either it is turned off or and interferometer error occured."
            )
        self.set_with_feedback_disabled(val)

    def set_with_feedback_disabled(self, val):
        target_val = val * self.parent.sign
        self.setpoint = target_val
        self.controller.socket_put(f"pa{self.parent.axis_Id_numeric},{target_val:.4f}")


class RtLamniMotorIsMoving(RtLamniSignalRO):
    def _socket_get(self):
        return self.controller.is_axis_moving(self.parent.axis_Id_numeric)

    def get(self):
        val = super().get()
        if val is not None:
            self._run_subs(
                sub_type=self.SUB_VALUE,
                value=val,
                timestamp=time.time(),
            )
        return val


class RtLamniFeedbackRunning(RtLamniSignalRO):
    @threadlocked
    def _socket_get(self):
        if int((self.controller.socket_put_and_receive("J2")).split(",")[0]) == 0:
            return 1
        else:
            return 0


class RtLamniMotor(Device, PositionerBase):
    USER_ACCESS = ["controller"]
    readback = Cpt(
        RtLamniReadbackSignal,
        signal_name="readback",
        kind="hinted",
    )
    user_setpoint = Cpt(RtLamniSetpointSignal, signal_name="setpoint")

    motor_is_moving = Cpt(RtLamniMotorIsMoving, signal_name="motor_is_moving", kind="normal")
    high_limit_travel = Cpt(Signal, value=0, kind="omitted")
    low_limit_travel = Cpt(Signal, value=0, kind="omitted")

    SUB_READBACK = "readback"
    SUB_CONNECTION_CHANGE = "connection_change"
    _default_sub = SUB_READBACK

    def __init__(
        self,
        axis_Id,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        host="mpc2680.psi.ch",
        port=3333,
        sign=1,
        socket_cls=SocketIO,
        device_manager=None,
        limits=None,
        **kwargs,
    ):
        self.axis_Id = axis_Id
        self.sign = sign
        self.controller = RtLamniController(socket=socket_cls(host=host, port=port))
        self.controller.set_axis(axis=self, axis_nr=self.axis_Id_numeric)
        self.device_manager = device_manager
        self.tolerance = kwargs.pop("tolerance", 0.5)

        super().__init__(
            prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs,
        )
        self.readback.name = self.name
        self.controller.subscribe(
            self._update_connection_state, event_type=self.SUB_CONNECTION_CHANGE
        )
        self._update_connection_state()

        # self.readback.subscribe(self._forward_readback, event_type=self.readback.SUB_VALUE)
        if limits is not None:
            assert len(limits) == 2
            self.low_limit_travel.put(limits[0])
            self.high_limit_travel.put(limits[1])

    @property
    def limits(self):
        return (self.low_limit_travel.get(), self.high_limit_travel.get())

    @property
    def low_limit(self):
        return self.limits[0]

    @property
    def high_limit(self):
        return self.limits[1]

    def check_value(self, pos):
        """Check that the position is within the soft limits"""
        low_limit, high_limit = self.limits

        if low_limit < high_limit and not (low_limit <= pos <= high_limit):
            raise LimitError(f"position={pos} not within limits {self.limits}")

    def _update_connection_state(self, **kwargs):
        for walk in self.walk_signals():
            walk.item._metadata["connected"] = self.controller.connected

    def _forward_readback(self, **kwargs):
        kwargs.pop("sub_type")
        self._run_subs(sub_type="readback", **kwargs)

    @raise_if_disconnected
    def move(self, position, wait=True, **kwargs):
        """Move to a specified position, optionally waiting for motion to
        complete.

        Parameters
        ----------
        position
            Position to move to
        moved_cb : callable
            Call this callback when movement has finished. This callback must
            accept one keyword argument: 'obj' which will be set to this
            positioner instance.
        timeout : float, optional
            Maximum time to wait for the motion. If None, the default timeout
            for this positioner is used.

        Returns
        -------
        status : MoveStatus

        Raises
        ------
        TimeoutError
            When motion takes longer than `timeout`
        ValueError
            On invalid positions
        RuntimeError
            If motion fails other than timing out
        """
        self._started_moving = False
        timeout = kwargs.pop("timeout", 100)
        status = super().move(position, timeout=timeout, **kwargs)
        self.user_setpoint.put(position, wait=False)

        def move_and_finish():
            while self.motor_is_moving.get():
                print("motor is moving")
                val = self.readback.read()
                self._run_subs(
                    sub_type=self.SUB_READBACK,
                    value=val,
                    timestamp=time.time(),
                )
                time.sleep(0.01)
            print("Move finished")
            self._done_moving()

        threading.Thread(target=move_and_finish, daemon=True).start()
        try:
            if wait:
                status_wait(status)
        except KeyboardInterrupt:
            self.stop()
            raise

        return status

    @property
    def axis_Id(self):
        return self._axis_Id_alpha

    @axis_Id.setter
    def axis_Id(self, val):
        if isinstance(val, str):
            if len(val) != 1:
                raise ValueError(f"Only single-character axis_Ids are supported.")
            self._axis_Id_alpha = val
            self._axis_Id_numeric = ord(val.lower()) - 97
        else:
            raise TypeError(f"Expected value of type str but received {type(val)}")

    @property
    def axis_Id_numeric(self):
        return self._axis_Id_numeric

    @axis_Id_numeric.setter
    def axis_Id_numeric(self, val):
        if isinstance(val, int):
            if val > 26:
                raise ValueError(f"Numeric value exceeds supported range.")
            self._axis_Id_alpha = val
            self._axis_Id_numeric = (chr(val + 97)).capitalize()
        else:
            raise TypeError(f"Expected value of type int but received {type(val)}")

    def kickoff(self, metadata, **kwargs) -> None:
        self.controller.kickoff(metadata)

    @property
    def egu(self):
        """The engineering units (EGU) for positions"""
        return "um"

    # how is this used later?

    def stage(self) -> list[object]:
        return super().stage()

    def unstage(self) -> list[object]:
        return super().unstage()

    def stop(self, *, success=False):
        self.controller.stop_all_axes()
        return super().stop(success=success)


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)

    mock = False
    if not mock:
        rty = RtLamniMotor("B", name="rty", host="mpc2680.psi.ch", port=3333, sign=1)
        rty.stage()
        status = rty.move(0, wait=True)
        status = rty.move(10, wait=True)
        rty.read()

        rty.get()
        rty.describe()

        rty.unstage()
    else:
        from ophyd_devices.utils.socket import SocketMock

        rtx = RtLamniMotor("A", name="rtx", host="mpc2680.psi.ch", port=3333, socket_cls=SocketMock)
        rty = RtLamniMotor("B", name="rty", host="mpc2680.psi.ch", port=3333, socket_cls=SocketMock)
        rtx.stage()
        # rty.stage()



import json

from ophyd import Component as Cpt
from ophyd import Device, PositionerBase, Signal
from ophyd.ophydobj import OphydObject
from ophyd_devices.utils.socket import SocketMock


def get_user_functions(obj) -> list:
    exclude_list = ["log", "SUB_CONNECTION_CHANGE"]
    exclude_classes = [Device, OphydObject, PositionerBase, Signal, Cpt]
    for cls in exclude_classes:
        exclude_list.extend(dir(cls))
    access_list = [
        func for func in dir(obj) if not func.startswith("_") and func not in exclude_list
    ]

    return access_list


def is_serializable(f) -> bool:
    try:
        json.dumps(f)
        return True
    except (TypeError, OverflowError):
        return False


def get_user_interface(obj, obj_interface):
    # user_funcs = get_user_functions(obj)
    for f in [f for f in dir(obj) if f in obj.USER_ACCESS]:
        if f == "controller" or f == "on":
            print(f)
        m = getattr(obj, f)
        if not callable(m):
            if is_serializable(m):
                obj_interface[f] = {"type": type(m).__name__}
            elif isinstance(m, SocketMock):
                pass
            else:
                obj_interface[f] = get_user_interface(m, {})
        else:
            obj_interface[f] = {"type": "func"}
    return obj_interface



from ophyd import Component as Cpt
from ophyd import Device, EpicsSignal


class Xeye(Device):
    save_frame = Cpt(EpicsSignal, "XOMNY-XEYE-SAVEFRAME:0")
    acquisition_done = Cpt(EpicsSignal, "XOMNY-XEYE-ACQDONE:0")
    acquisition = Cpt(EpicsSignal, "XOMNY-XEYE-ACQ:0")
    x_width = Cpt(EpicsSignal, "XOMNY-XEYE-XWIDTH_X:0")



import time as ttime

from bec_lib import bec_logger
from ophyd import Signal
from ophyd.utils import ReadOnlyError

logger = bec_logger.logger


class ReadbackSignal(Signal):
    """Readback signal for simulated devices.

    It will return the value of the readback signal based on the position
    created in the sim_state dictionary of the parent device.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._metadata.update(
            connected=True,
            write_access=False,
        )

    def get(self, **kwargs):
        """Get the current position of the simulated device."""
        self._readback = self.parent.sim_state["readback"]
        self.parent.sim_state["readback_ts"] = ttime.time()
        return self._readback

    def describe(self):
        """Describe the readback signal."""
        res = super().describe()
        res[self.name]["precision"] = self.parent.precision
        return res

    @property
    def timestamp(self):
        """Timestamp of the readback value"""
        return self.parent.sim_state["readback_ts"]

    def put(self, value, *, timestamp=None, force=False, **kwargs):
        """Put method, should raise ReadOnlyError since the signal is readonly."""
        raise ReadOnlyError(f"The signal {self.name} is readonly.")

    def set(self, value, *, timestamp=None, force=False, **kwargs):
        """Set method, should raise ReadOnlyError since the signal is readonly."""
        raise ReadOnlyError(f"The signal {self.name} is readonly.")


class SetpointSignal(Signal):
    """Setpoint signal for simulated devices.

    When read, it will return the "setpoint" key from the dictionary sim_state,
    and whe put it will call the set method of the parent device with the value.
    """

    def put(self, value, *, timestamp=None, force=False, **kwargs):
        """Put the value to the simulated device."""
        self._readback = float(value)
        self.parent.set(float(value))

    def get(self, **kwargs):
        """Get the current setpoint of the simulated device."""
        self._readback = self.parent.sim_state["setpoint"]
        return self.parent.sim_state["setpoint"]

    @property
    def timestamp(self):
        """Timestamp of the readback value"""
        return self.parent.sim_state["setpoint_ts"]


class IsMovingSignal(Signal):
    """IsMoving signal for simulated devices.

    When read, it will return the "is_moving" key from the dictionary sim_state,
    and whe put it will call the set method of the parent device with the value.
    """

    def get(self, **kwargs):
        self._readback = self.parent.sim_state["is_moving"]
        self.parent.sim_state["is_moving_ts"] = ttime.time()
        return self.parent.sim_state["is_moving"]

    @property
    def timestamp(self):
        """Timestamp of the readback value"""
        return self.parent.sim_state["is_moving_ts"]

    def put(self, value, *, timestamp=None, force=False, **kwargs):
        """Put method, should raise ReadOnlyError since the signal is readonly."""
        raise ReadOnlyError(f"The signal {self.name} is readonly.")

    def set(self, value, *, timestamp=None, force=False, **kwargs):
        """Set method, should raise ReadOnlyError since the signal is readonly."""
        raise ReadOnlyError(f"The signal {self.name} is readonly.")
